import os
import sys
import tempfile
import traceback
import json
import warnings
import pickle
import math
from pathlib import Path

HAS_FASTAI = True
import_exception = None

import arcgis
from arcgis.features import FeatureLayer

try:
    from ._arcgis_model import ArcGISModel, _raise_fastai_import_error
    from fastai.basic_train import Learner, load_learner
    from fastprogress.fastprogress import progress_bar
    from .._utils.tabular_data import TabularDataObject
    from fastai.torch_core import split_model_idx
    import torch
    from fastai.metrics import r2_score
    from sklearn.preprocessing import LabelEncoder
    from ._tsmodel_archs._InceptionTime import _TSInceptionTime
    from ._tsmodel_archs._Resnet import _TSResNet
    from ._tsmodel_archs._ResCNN import _TSResCNN
    from ._tsmodel_archs._FCN import _TSFCN
    from ._tsmodel_archs._LSTM import _TSLSTM
    from .._utils.TSData import To3dTensor, ToTensor
    from .._utils.common import _get_emd_path
    from ._tsmodel_archs._TST import TST

    _model_arch = {
        "inceptiontime": _TSInceptionTime,
        "resnet": _TSResNet,
        "rescnn": _TSResCNN,
        "fcn": _TSFCN,
        "lstm": _TSLSTM,
        "timeseriestransformer": TST,
    }

except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False
    _model_arch = {}

HAS_NUMPY = True
try:
    import numpy as np
except:
    HAS_NUMPY = False

_PROTOCOL_LEVEL = 2

HAS_PANDAS = True
try:
    import pandas as pd
except:
    HAS_PANDAS = False


def _get_model_from_path(pretrained_path):
    learn = load_learner(
        os.path.dirname(pretrained_path),
        os.path.basename(pretrained_path).rsplit(".", 1)[0] + "_exported.pth",
        no_check=True,
    )

    return learn


class TimeSeriesModel(ArcGISModel):
    """
    Creates a :class:`~arcgis.learn.TimeSeriesModel` Object.
    Based on the Fast.ai's https://github.com/timeseriesAI/timeseriesAI

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required TabularDataObject. Returned data object from
                            :class:`~arcgis.learn.prepare_tabulardata` function.
    ---------------------   -------------------------------------------
    seq_len                 Required Integer. Sequence Length for the series.
                            In case of raster only, seq_len = number of rasters,
                            any other passed value will be ignored.
    ---------------------   -------------------------------------------
    model_arch              Optional string. Model Architecture.
                            Allowed "InceptionTime", "ResCNN",
                            "Resnet", "FCN", "TimeSeriesTransformer", "LSTM". "LSTM"
                            supports both "LSTM" and "Bi-LSTM". "Bi-LSTM" is enabled by passing
                            `bidirectional=True` in kwargs.
    ---------------------   -------------------------------------------
    location_var            Optional string. Location variable in case of
                            NetCDF dataset.
    ---------------------   -------------------------------------------
    multistep               Optional string. It will set the model to generate
                            more than one time-step as output in multivariate scenario.
                            Compared to current auto-regressive fashion, it will generate
                            multi-step output in single pass.
                            This option is only  applicable in multivariate
                            scenario. Univariate implementation will ignore this flag.
                            Default value is `False`
    ---------------------   -------------------------------------------
    ``**kwargs``            Optional kwargs.
    ---------------------   -------------------------------------------
    =====================   ===========================================

    :return: :class:`~arcgis.learn.TimeSeriesModel` Object
    """

    def __init__(
        self,
        data,
        seq_len,
        model_arch="InceptionTime",
        location_var=None,
        multistep=False,
        **kwargs,
    ):
        data_bunch = None
        if not data._is_empty:
            data_bunch = data._time_series_bunch(
                seq_len, location_var, multistep=multistep
            )

        super().__init__(data, None)
        self.multistep = multistep
        # We need fix the number of steps based on whether it is univariate or multivariate. We are not supporting
        # multi-step training for univariate case.
        fields_needed = (
            data._categorical_variables
            + data._continuous_variables
            + data._dependent_variable
        )
        self.step = 1
        if self.multistep and len(list(fields_needed)) != 1:
            self.step = seq_len // 2

        if not data_bunch:
            self.learn = _get_model_from_path(kwargs.get("pretrained_path"))
        elif kwargs.get("pretrained_path"):
            self.learn = _get_model_from_path(kwargs.get("pretrained_path"))
            self.learn.data = data_bunch
        else:
            if not _model_arch.get(model_arch.lower()):
                raise Exception("Invalid model architecture")

            model_arch_ob = _model_arch.get(model_arch.lower())
            if model_arch.lower() == "lstm":
                model = model_arch_ob(
                    data_bunch.features, data_bunch.c, self._device, **kwargs
                ).to(self._device)
            elif model_arch.lower() == "timeseriestransformer":
                model = model_arch_ob(
                    data_bunch.features, data_bunch.c, seq_len, **kwargs
                ).to(self._device)
            else:
                if model_arch.lower() in ["resnet", "fcn"]:
                    kwargs["device"] = self._device
                model = model_arch_ob(data_bunch.features, data_bunch.c, **kwargs).to(
                    self._device
                )
                if model_arch.lower() in ["resnet", "fcn"]:
                    del kwargs["device"]
            self.learn = Learner(data_bunch, model, path=data.path)
            self.learn.data = data_bunch

        self.learn.layer_groups = split_model_idx(self.learn.model, [1])
        self._model_arch = model_arch.lower()
        if kwargs.get("pretrained_path"):
            del kwargs["pretrained_path"]
        self._kwargs = kwargs
        self._seq_len = seq_len

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.TimeSeriesModel` Object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :class:`~arcgis.learn.prepare_tabulardata` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.TimeSeriesModel` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        dependent_variable = emd["dependent_variable"]
        # added reverse support to the EMD params.
        if isinstance(dependent_variable, str):
            dependent_variable = [dependent_variable]
        categorical_variables = emd["categorical_variables"]
        continuous_variables = emd["continuous_variables"]

        _is_classification = False
        if emd["_is_classification"] == "classification":
            _is_classification = True

        model_params = emd["model_params"]
        model_arch = emd["model_arch"]
        seq_len = emd["seq_len"]
        index_field = emd.get("index_field", None)
        test_size = emd.get("test_size", None)
        step = emd.get("step", 1)
        multistep = emd.get("multistep", False)
        # encoder_path = os.path.join(os.path.dirname(emd_path),
        #                             os.path.basename(emd_path).split('.')[0] + '_encoders.pkl')

        scaler_path = os.path.join(
            os.path.dirname(emd_path),
            os.path.basename(emd_path).split(".")[0] + "_scaler.pkl",
        )

        encoder_mapping = None
        scaler = None

        # if os.path.exists(encoder_path):
        #     with open(encoder_path, 'rb') as f:
        #         encoder_mapping = pickle.loads(f.read())

        if os.path.exists(scaler_path):
            with open(scaler_path, "rb") as f:
                scaler = pickle.loads(f.read())

        if data is None:
            data = TabularDataObject._empty(
                categorical_variables,
                continuous_variables,
                dependent_variable,
                encoder_mapping,
            )
            data._is_classification = _is_classification
            data._column_transforms_mapping = scaler

            # if index_field is not None:
            data._index_field = index_field
            data._test_size = test_size

            class_object = cls(
                data,
                seq_len,
                model_arch=model_arch,
                pretrained_path=emd_path,
                step=step,
                multistep=multistep,
                **model_params,
            )
            class_object._data.emd = emd
            class_object._data.emd_path = emd_path
            return class_object

        return cls(
            data,
            seq_len,
            model_arch=model_arch,
            pretrained_path=emd_path,
            step=step,
            multistep=multistep,
            **model_params,
        )

    def save(
        self,
        name_or_path,
        framework="PyTorch",
        publish=False,
        gis=None,
        save_optimizer=False,
        **kwargs,
    ):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by :class:`~arcgis.learn.SingleShotDetector`, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` can be
                                passed as an optional keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
        =====================   ===========================================
        """

        if "\\" in name_or_path or "/" in name_or_path:
            path = os.path.abspath(name_or_path)
        else:
            path = os.path.join(self._data.path, "models", name_or_path)
            if not os.path.exists(os.path.dirname(path)):
                os.mkdir(os.path.dirname(path))

        if not os.path.exists(path):
            os.mkdir(path)

        base_file_name = os.path.basename(path)

        TimeSeriesModel._save_encoders(
            self._data._column_transforms_mapping, path, base_file_name
        )
        self.learn.export(os.path.join(path, os.path.basename(path) + "_exported.pth"))
        from IPython.utils import io

        with io.capture_output() as captured:
            saved_path = super().save(
                path, framework, False, gis, save_optimizer=save_optimizer, **kwargs
            )
        if publish:
            file_name = os.path.basename(saved_path) + ".dlpk"
            dlpk_path = Path(os.path.join(saved_path, file_name))
            self._publish_dlpk(
                dlpk_path,
                gis=gis,
                overwrite=kwargs.get("overwrite", False),
            )

        return Path(path)

    @staticmethod
    def _save_encoders(scaler, path, base_file_name):
        # if not encoder_mapping:
        #     return

        # encoder_file = os.path.join(path, base_file_name + '_encoders.pkl')
        # with open(encoder_file, 'wb') as f:
        #     f.write(pickle.dumps(encoder_mapping, protocol=_PROTOCOL_LEVEL))

        scaler_file = os.path.join(path, base_file_name + "_scaler.pkl")
        with open(scaler_file, "wb") as f:
            f.write(pickle.dumps(scaler, protocol=_PROTOCOL_LEVEL))

    @property
    def _model_metrics(self):
        # from IPython.utils import io
        # with io.capture_output() as captured:
        #     score = self.score()

        return {"score": self.score()}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["ModelType"] = "TimeSeriesModel"
        _emd_template["model_arch"] = self._model_arch
        _emd_template["model_params"] = self._kwargs
        _emd_template["seq_len"] = self._seq_len
        # Added this for backward compatibility.
        # Added the logic to enable compatibility with current ArcgisPro
        if len(self._data._dependent_variable) == 1:
            _emd_template["dependent_variable"] = self._data._dependent_variable[0]
        else:
            _emd_template["dependent_variable"] = self._data._dependent_variable
        _emd_template["categorical_variables"] = self._data._categorical_variables
        _emd_template["continuous_variables"] = self._data._continuous_variables
        _emd_template["multistep"] = self.multistep
        _emd_template["step"] = self.step

        if self._data._index_field:
            _emd_template["index_field"] = self._data._index_field

        _emd_template["_is_classification"] = (
            "classification" if self._data._is_classification else "regression"
        )
        if hasattr(self._data, "_test_size"):
            _emd_template["test_size"] = self._data._test_size

        return _emd_template

    def _predict(self, orig_sequence):
        sequence = np.array(orig_sequence, dtype="float32")
        if sequence.shape[0] == 1:
            seq_arr = To3dTensor(orig_sequence).to(self._device)
        else:
            seq_arr = ToTensor(
                np.expand_dims(np.array(sequence, dtype="float32"), axis=0)
            ).to(self._device)
        model = self.learn.model
        model.eval()

        with torch.no_grad():
            output = model(seq_arr).cpu().numpy()

        return output

    def predict(
        self,
        input_features=None,
        explanatory_rasters=None,
        datefield=None,
        distance_features=None,
        output_layer_name="Prediction Layer",
        gis=None,
        prediction_type="features",
        output_raster_path=None,
        match_field_names=None,
        number_of_predictions=None,
    ):
        """

        Predict on data from feature layer and or raster data.

        =================================   =========================================================================
        **Parameter**                        **Description**
        ---------------------------------   -------------------------------------------------------------------------
        input_features                      Optional :class:`~arcgis.features.FeatureLayer` or spatially enabled dataframe.
                                            Contains features with location of the input data.
                                            Required if prediction_type is 'features' or 'dataframe'
        ---------------------------------   -------------------------------------------------------------------------
        explanatory_rasters                 Optional list of Raster Objects.
                                            Required if prediction_type is 'rasters'
        ---------------------------------   -------------------------------------------------------------------------
        datefield                           Optional field_name.
                                            This field contains the date in the input_features.
                                            The field type can be a string or date time field.
                                            If specified, the field will be split into
                                            Year, month, week, day, dayofweek, dayofyear,
                                            is_month_end, is_month_start, is_quarter_end,
                                            is_quarter_start, is_year_end, is_year_start,
                                            hour, minute, second, elapsed and these will be added
                                            to the prepared data as columns.
                                            All fields other than elapsed and dayofyear are treated
                                            as categorical.
        ---------------------------------   -------------------------------------------------------------------------
        distance_features                   Optional List of :class:`~arcgis.features.FeatureLayer` objects.
                                            These layers are used for calculation of field "NEAR_DIST_1",
                                            "NEAR_DIST_2" etc in the output dataframe.
                                            These fields contain the nearest feature distance
                                            from the input_features.
                                            Same as :meth:`~arcgis.learn.prepare_tabulardata`.
        ---------------------------------   -------------------------------------------------------------------------
        output_layer_name                   Optional string. Used for publishing the output layer.
        ---------------------------------   -------------------------------------------------------------------------
        gis                                 Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                            If not specified then active gis user is taken.
        ---------------------------------   -------------------------------------------------------------------------
        prediction_type                     Optional String.
                                            Set 'features' or 'dataframe' to make output predictions.
        ---------------------------------   -------------------------------------------------------------------------
        output_raster_path                  Optional path. Required when prediction_type='raster', saves
                                            the output raster to this path.
        ---------------------------------   -------------------------------------------------------------------------
        match_field_names                   Optional string.
                                            Specify mapping of the original training set with prediction set.
        ---------------------------------   -------------------------------------------------------------------------
        number_of_predictions               Optional int for univariate time series.
                                            Specify the number of predictions to make, adds new rows to the dataframe.
                                            For multivariate or if None, it expects the dataframe to have empty rows.
                                            if multi-step is set to True during training then it does not need empty
                                            rows. If multi-step is set to False then dataframe needs to have rows with
                                            `NA` values in `variable predict` and non-NA values in `explnatory_varibles`
                                            For prediction_type='raster', a new raster is created.
        =================================   =========================================================================

        :return: :class:`~arcgis.features.FeatureLayer`/dataframe if prediction_type='features'/'dataframe', else returns True and saves output
        raster at the specified path.
        """

        rasters = explanatory_rasters if explanatory_rasters else []
        if prediction_type in ["features", "dataframe"]:
            if input_features is None:
                raise Exception("Feature Layer required for predict_features=True")

            gis = gis if gis else arcgis.env.active_gis
            return self._predict_features(
                input_features,
                rasters,
                datefield,
                distance_features,
                output_layer_name,
                gis,
                match_field_names,
                number_of_predictions,
                prediction_type,
            )
        else:
            if not rasters:
                raise Exception("Rasters required for predict_features=False")

            if not output_raster_path:
                raise Exception(
                    "Please specify output_raster_folder_path to save the output."
                )

            return self._predict_rasters(output_raster_path, rasters, match_field_names)

    def _predict_rasters(self, output_raster_path, rasters, match_field_names=None):
        if not os.path.exists(os.path.dirname(output_raster_path)):
            raise Exception("Output directory doesn't exist")

        if os.path.exists(output_raster_path):
            raise Exception("Output Folder already exists")

        try:
            import arcpy
        except:
            raise Exception("This function requires arcpy.")

        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")

        fields_needed = (
            self._data._categorical_variables + self._data._continuous_variables
        )

        try:
            arcpy.env.outputCoordinateSystem = rasters[0].extent["spatialReference"][
                "wkt"
            ]
        except:
            arcpy.env.outputCoordinateSystem = rasters[0].extent["spatialReference"][
                "wkid"
            ]

        xmin = rasters[0].extent["xmin"]
        xmax = rasters[0].extent["xmax"]
        ymin = rasters[0].extent["ymin"]
        ymax = rasters[0].extent["ymax"]
        min_cell_size_x = rasters[0].mean_cell_width
        min_cell_size_y = rasters[0].mean_cell_height

        default_sr = rasters[0].extent["spatialReference"]

        for raster in rasters:
            point_upper = arcgis.geometry.Point(
                {
                    "x": raster.extent["xmin"],
                    "y": raster.extent["ymax"],
                    "sr": raster.extent["spatialReference"],
                }
            )
            point_lower = arcgis.geometry.Point(
                {
                    "x": raster.extent["xmax"],
                    "y": raster.extent["ymin"],
                    "sr": raster.extent["spatialReference"],
                }
            )
            cell_size = arcgis.geometry.Point(
                {
                    "x": raster.mean_cell_width,
                    "y": raster.mean_cell_height,
                    "sr": raster.extent["spatialReference"],
                }
            )

            points = arcgis.geometry.project(
                [point_upper, point_lower, cell_size],
                raster.extent["spatialReference"],
                default_sr,
            )
            point_upper = points[0]
            point_lower = points[1]
            cell_size = points[2]

            if xmin > point_upper.x:
                xmin = point_upper.x
            if ymax < point_upper.y:
                ymax = point_upper.y
            if xmax < point_lower.x:
                xmax = point_lower.x
            if ymin > point_lower.y:
                ymin = point_lower.y

            if min_cell_size_x > cell_size.x:
                min_cell_size_x = cell_size.x

            if min_cell_size_y > cell_size.y:
                min_cell_size_y = cell_size.y

        max_raster_columns = int(abs(math.ceil((xmax - xmin) / min_cell_size_x)))
        max_raster_rows = int(abs(math.ceil((ymax - ymin) / min_cell_size_y)))

        point_upper = arcgis.geometry.Point({"x": xmin, "y": ymax, "sr": default_sr})
        cell_size = arcgis.geometry.Point(
            {"x": min_cell_size_x, "y": min_cell_size_y, "sr": default_sr}
        )

        raster_data = {}
        for raster in rasters:
            field_name = raster.name
            point_upper_translated = arcgis.geometry.project(
                [point_upper], default_sr, raster.extent["spatialReference"]
            )[0]
            cell_size_translated = arcgis.geometry.project(
                [cell_size], default_sr, raster.extent["spatialReference"]
            )[0]
            if field_name in fields_needed:
                raster_read = raster.read(
                    origin_coordinate=(
                        point_upper_translated.x,
                        point_upper_translated.y,
                    ),
                    ncols=max_raster_columns,
                    nrows=max_raster_rows,
                    cell_size=(cell_size_translated.x, cell_size_translated.y),
                )
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f"_{index}"
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            elif match_field_names and match_field_names.get(raster.name):
                field_name = match_field_names.get(raster.name)
                raster_read = raster.read(
                    origin_coordinate=(
                        point_upper_translated.x,
                        point_upper_translated.y,
                    ),
                    ncols=max_raster_columns,
                    nrows=max_raster_rows,
                    cell_size=(cell_size_translated.x, cell_size_translated.y),
                )
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f"_{index}"
                            if not raster_data.get(key):
                                raster_data[key] = []

                            index = index + 1
                            raster_data[key].append(value)
            else:
                continue

        for field in fields_needed:
            if (
                field not in list(raster_data.keys())
                and match_field_names
                and match_field_names.get(field, None) is None
            ):
                raise Exception(f"Field missing {field}")

        length_values = len(raster_data[list(raster_data.keys())[0]])
        processed_output = []
        for i in progress_bar(range(length_values)):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_output.append(self._predict([processed_row]))

        processed_numpy = np.array(processed_output, dtype="float64")
        processed_numpy = processed_numpy.reshape([max_raster_rows, max_raster_columns])
        processed_raster = arcpy.NumPyArrayToRaster(
            processed_numpy,
            arcpy.Point(xmin, ymin),
            x_cell_size=min_cell_size_x,
            y_cell_size=min_cell_size_y,
        )
        processed_raster.save(output_raster_path)

        return True

    def _predict_features(
        self,
        input_features,
        rasters=None,
        datefield=None,
        distance_features=None,
        output_layer_name="Prediction Layer",
        gis=None,
        match_field_names=None,
        number_of_predictions=None,
        prediction_type="features",
    ):
        if not HAS_PANDAS:
            raise Exception("This function requires pandas library")

        if isinstance(input_features, FeatureLayer):
            orig_dataframe = input_features.query().sdf
        else:
            orig_dataframe = input_features.copy()

        # Dtype conversion because native pandas format will break the plotting libraries
        for i in orig_dataframe.columns:
            if isinstance(orig_dataframe.loc[:, i].dtype, pd.Float64Dtype):
                orig_dataframe.loc[:, i] = orig_dataframe.loc[:, i].astype(np.float64)

        if match_field_names is None:
            match_field_names = {}
        fields_needed = (
            self._data._categorical_variables
            + self._data._continuous_variables
            + self._data._dependent_variable
        )

        (
            orig_dataframe,
            single_swap_pred,
            number_of_predictions,
        ) = self._infer_number_of_pred(
            orig_dataframe, number_of_predictions, match_field_names, fields_needed
        )

        dataframe = orig_dataframe.copy()
        distance_feature_layers = distance_features if distance_features else []

        continuous_variables = (
            self._data._continuous_variables + self._data._dependent_variable
        )

        feature_layer_columns = []
        for column in dataframe.columns:
            column_name = column
            categorical = False
            if column_name in fields_needed:
                if column_name not in continuous_variables:
                    categorical = True
            elif match_field_names and match_field_names.get(column_name):
                if match_field_names.get(column_name) not in continuous_variables:
                    categorical = True
            else:
                continue

            feature_layer_columns.append((column_name, categorical))

        raster_columns = []
        if rasters:
            for raster in rasters:
                column_name = raster.name
                categorical = False
                if column_name in fields_needed:
                    if column_name not in continuous_variables:
                        categorical = True
                elif match_field_names and match_field_names.get(column_name):
                    column_name = match_field_names.get(column_name)
                    if column_name not in continuous_variables:
                        categorical = True
                else:
                    continue

                raster_columns.append((raster, categorical))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            (
                processed_dataframe,
                fields_mapping,
            ) = TabularDataObject._prepare_dataframe_from_features(
                orig_dataframe,
                self._data._dependent_variable,
                feature_layer_columns,
                raster_columns,
                datefield,
                distance_feature_layers,
            )

        if match_field_names:
            processed_dataframe.rename(columns=match_field_names, inplace=True)

        for field in fields_needed:
            if field not in processed_dataframe.columns:
                raise Exception(f"Field missing {field}")

        for column in processed_dataframe.columns:
            if column not in fields_needed:
                processed_dataframe = processed_dataframe.drop(column, axis=1)

        # preserve the ordering of the column, It is required because sorting tends to change the order and ultimately
        # affects different transforms.

        order_columns = (
            self._data._dependent_variable
            + self._data._continuous_variables
            + self._data._categorical_variables
        )

        processed_dataframe = processed_dataframe.loc[:, order_columns]

        # processed_dataframe = processed_dataframe.reindex(
        #     sorted(processed_dataframe.columns), axis=1
        # )

        index = self._seq_len
        processed_dataframe[self._data._dependent_variable] = processed_dataframe[
            self._data._dependent_variable
        ].replace(r"^\s*$", np.nan, regex=True)

        processed_dataframe_transform = processed_dataframe.copy()

        processed_dataframe_transform = self._apply_transform(
            processed_dataframe, processed_dataframe_transform
        )

        big_bunch = []
        prediction_sequence_list = []
        processed_dataframe_transform = processed_dataframe_transform.values
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            for col in range(len(processed_dataframe.columns.values)):
                if (
                    list(processed_dataframe.columns.values)[col]
                    in self._data._dependent_variable
                ):
                    prediction_sequence_list.append(
                        processed_dataframe_transform[:, col]
                    )
                    ind = len(prediction_sequence_list) - 1
                    big_bunch.append(prediction_sequence_list[ind][0 : self._seq_len])
                else:
                    big_bunch.append(
                        processed_dataframe_transform[:, col][0 : self._seq_len]
                    )
        # checking the first array because the changes would force this to be of type List[List]
        prediction_sequence_list = np.stack(prediction_sequence_list, axis=1)
        if prediction_sequence_list.shape[0] < self._seq_len:
            raise Exception("Basic Sequence not found!")
        # modified below column so that multivariates can be captured
        while index < len(prediction_sequence_list):
            if pd.isna(prediction_sequence_list[index]).any() or any(
                [
                    True
                    if i
                    in [
                        "",
                        None,
                        "null",
                        "None",
                    ]
                    else False
                    for i in prediction_sequence_list[index]
                ]
            ):
                value = self._predict(np.array(big_bunch))
                # Now the predicted value will have the shape of 1 * [self.step * len(variable_predict)
                if not single_swap_pred:
                    prediction_sequence_list[index] = value.reshape(self.step, -1).mean(
                        axis=0
                    )
                else:
                    prediction_sequence_list[index:] = value.reshape(self.step, -1)

            index = index + 1
            big_bunch = []
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                for col in range(len(processed_dataframe.columns.values)):
                    if (
                        list(processed_dataframe.columns.values)[col]
                        in self._data._dependent_variable
                    ):
                        idx = self._data._dependent_variable.index(
                            list(processed_dataframe.columns.values)[col]
                        )
                        big_bunch.append(
                            prediction_sequence_list[:, idx][
                                index
                                - self._seq_len : (
                                    index - self._seq_len + self._seq_len
                                )
                            ]
                        )
                    else:
                        big_bunch.append(
                            processed_dataframe_transform[:, col][
                                index
                                - self._seq_len : (
                                    index - self._seq_len + self._seq_len
                                )
                            ]
                        )
        transformed_results = prediction_sequence_list
        transformed_results_col = self._apply_inverse_transform(transformed_results)

        for idx, col in enumerate(self._data._dependent_variable):
            orig_dataframe[col + "_results"] = transformed_results_col[:, idx]

        if prediction_type == "dataframe":
            return orig_dataframe

        if "SHAPE" in list(orig_dataframe.columns):
            orig_dataframe.spatial.to_featurelayer(output_layer_name, gis)
        else:
            import tempfile

            with tempfile.TemporaryDirectory() as tmpdir:
                table_file = os.path.join(tmpdir, output_layer_name + ".xlsx")
                orig_dataframe.to_excel(table_file, index=False, header=True)
                online_table = gis.content.add(
                    {"type": "Microsoft Excel", "overwrite": True}, table_file
                )
                return online_table.publish(overwrite=True)

    def _apply_inverse_transform(self, transformed_results):
        # Apply inverse transform to generate the main data
        transformed_results_ret = []
        for idx, col in enumerate(self._data._dependent_variable):
            if self._data._column_transforms_mapping.get(col):
                for transform in self._data._column_transforms_mapping.get(col):
                    transformed_results_col = transform.inverse_transform(
                        np.array(transformed_results[:, idx]).reshape(-1, 1)
                    )
                    transformed_results_col = transformed_results_col.squeeze(1)
                transformed_results_ret.append(transformed_results_col)
        return np.stack(transformed_results_ret, axis=1)

    def _add_predict_rows(
        self, number_of_predictions, orig_dataframe, match_field_names, fields_needed
    ):
        # Changed to make code future ready as the previous method of adding
        # pandas series will be deprecated.
        #
        from pandas.api.types import is_datetime64_any_dtype as is_datetime

        datetime_dict = {}
        delta = None
        index_field_name = None
        end_value = None
        if self._data._index_field is not None:
            index_field_name = match_field_names.get(
                self._data._index_field, self._data._index_field
            )
            if index_field_name in list(orig_dataframe.columns):
                if is_datetime(orig_dataframe[index_field_name]):
                    delta = (
                        orig_dataframe[index_field_name].iloc[1]
                        - orig_dataframe[index_field_name].iloc[0]
                    )
                    end_value = None
                    if delta is not None:
                        end_value = orig_dataframe[index_field_name].iloc[
                            len(orig_dataframe) - 1
                        ]
                else:
                    orig_dataframe[index_field_name], sample = self._convert_datetime(
                        orig_dataframe[index_field_name]
                    )
                    if not sample:
                        delta = (
                            orig_dataframe[index_field_name].iloc[1]
                            - orig_dataframe[index_field_name].iloc[0]
                        )
                        end_value = None
                        if delta is not None:
                            end_value = orig_dataframe[index_field_name].iloc[
                                len(orig_dataframe) - 1
                            ]

        for i in orig_dataframe.columns:
            if i != index_field_name:
                if is_datetime(orig_dataframe[i]):
                    new_delta = orig_dataframe[i].iloc[1] - orig_dataframe[i].iloc[0]
                    end_value_temp = None
                    if new_delta is not None:
                        end_value_temp = orig_dataframe[i].iloc[len(orig_dataframe) - 1]
                    if new_delta is not None:
                        datetime_dict[i] = tuple([new_delta, end_value_temp])
                else:
                    if orig_dataframe[i].dtype == "object" and i in fields_needed:
                        # check whether it is time
                        orig_dataframe[i], sample = self._convert_datetime(
                            orig_dataframe[i]
                        )
                        if not sample:
                            new_delta = (
                                orig_dataframe[i].iloc[1] - orig_dataframe[i].iloc[0]
                            )
                            end_value_temp = None
                            if new_delta is not None:
                                end_value_temp = orig_dataframe[i].iloc[
                                    len(orig_dataframe) - 1
                                ]
                            if new_delta is not None:
                                datetime_dict[i] = tuple([new_delta, end_value_temp])

        pred_temp_df = pd.DataFrame(
            np.full([number_of_predictions, orig_dataframe.shape[1]], np.NAN)
        )
        pred_temp_df.columns = orig_dataframe.columns
        # preserve the indexes. Need to adjust 1 because new index will start from 0
        pred_temp_df.index += orig_dataframe.index[-1] + 1
        orig_dataframe = pd.concat([orig_dataframe, pred_temp_df])
        if delta is not None:
            tindex = pd.period_range(
                end_value, freq=delta, periods=number_of_predictions + 1
            )
            orig_dataframe.loc[
                orig_dataframe.tail(number_of_predictions).index, index_field_name
            ] = tindex[1:]
        if len(datetime_dict):
            for key, value in datetime_dict.items():
                new_delta, end_value_temp = value
                tindex = pd.period_range(
                    end_value_temp, freq=new_delta, periods=number_of_predictions + 1
                )
                orig_dataframe.loc[
                    orig_dataframe.tail(number_of_predictions).index, key
                ] = tindex[1:]

        return orig_dataframe

    def _infer_number_of_pred(
        self, orig_dataframe, number_of_predictions, match_field_names, fields_needed
    ):
        # Type of inference
        #     ├── Multivariate
        #     │   ├── MultiStep
        #     │   │   ├── NA rec in df>0 then num_of_pred= Number of NA rec/ otherwise warning
        #     │   │   └── Number_of_pred=0 or None and NA rec in df is 0 then seq_len//2
        #     │   └── Number of NA records in df will be honored
        #     │       └── Number_of_pred>0 then infer using NA in df warning
        #     └── Univariate
        #         └── Number_of_pred is honored
        ###

        single_swap_pred = False
        if self.multistep:
            df_na_count = len(orig_dataframe[orig_dataframe.isna().any(axis=1)])

            if df_na_count == 0:
                if not (number_of_predictions is None or number_of_predictions == 0):
                    warnings.warn(
                        f"Number of predictions is supplied in multivariate scenario. Overriding the value\
                                  with step value {self.step} "
                    )
                number_of_predictions = self.step
                single_swap_pred = True
                add_na_rec = True
            else:
                if not (number_of_predictions is None or number_of_predictions == 0):
                    warnings.warn(
                        f"Both Number of predictions and dataframe with NA is supplied.\
                                  Using dataframe NA count as number of predictions {df_na_count}"
                    )
                number_of_predictions = df_na_count
                add_na_rec = False
                warnings.warn(
                    "The model is trained with multistep objective. Setting NA values will result "
                    "in Auto-regressive mode of inference. This may lead to low accuracy."
                )

        else:
            if len(self._data._dependent_variable):
                number_of_predictions = number_of_predictions
                add_na_rec = True
            else:
                # multivariate case
                number_of_predictions = len(
                    orig_dataframe[orig_dataframe.isna().any(axis=1)]
                )
                add_na_rec = False

        if number_of_predictions is None:
            number_of_predictions = 0

        if add_na_rec:
            orig_dataframe = self._add_predict_rows(
                orig_dataframe=orig_dataframe,
                match_field_names=match_field_names,
                number_of_predictions=number_of_predictions,
                fields_needed=fields_needed,
            )
        return orig_dataframe, single_swap_pred, number_of_predictions

    def _apply_transform(self, processed_dataframe, processed_dataframe_transform):
        # changed to function because the transformation needs to handle the categorical variable case.
        # In such a scenario, the NaN is treated as label and label encoder fails because train data did not
        # have any NA label while the test data has because of nature of prediction.
        for col in list(processed_dataframe.columns):
            if col in self._data._categorical_variables:
                transformed_data = processed_dataframe[col].dropna()
            else:
                transformed_data = processed_dataframe[col]
            for transform in self._data._column_transforms_mapping.get(col, []):
                if isinstance(transform, LabelEncoder):
                    transformed_data = transform.transform(
                        np.array(
                            transformed_data,
                            dtype=type(processed_dataframe[col][0]),
                        )
                    )
                    transformed_data = transformed_data.reshape(-1, 1)
                else:
                    transformed_data = transform.transform(
                        np.array(
                            transformed_data,
                            dtype=type(processed_dataframe[col][0]),
                        ).reshape(-1, 1)
                    )

                transformed_data = transformed_data.squeeze(1)
            processed_dataframe_transform[col].head(len(transformed_data)).loc[
                :
            ] = np.array(transformed_data, dtype=type(processed_dataframe[col][0]))
        return processed_dataframe_transform

    def score(self):
        """
        :return: R2 score for regression model and Accuracy for classification model.
        """

        self._check_requisites()
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        model = self.learn.model

        model.eval()

        dl = self.learn.data.valid_dl

        targets = []
        predictions = []
        for i in range(len(dl.x.items)):
            prediction = self._predict(dl.x.items[i])[0]
            target = dl.y.items[i]
            # targets.append(target)
            if isinstance(target, (list, np.ndarray)):
                targets.append(target)
            else:
                targets.append([target])
            predictions.append(prediction)

        targets = np.array(targets, dtype="float64")
        predictions = np.array(predictions, dtype="float64")

        transformed_results = np.stack(targets, axis=0)
        transformed_results = self._apply_inverse_transform(transformed_results)
        targets_inversed = transformed_results

        transformed_results = np.stack(predictions, axis=0)

        transformed_results = self._apply_inverse_transform(transformed_results)
        predictions_inversed = transformed_results

        if self._data._is_classification:
            return (np.array(predictions) == np.array(targets)).mean()
        else:
            targets = torch.tensor(targets_inversed).to(self._device)
            predictions = torch.tensor(predictions_inversed).to(self._device)
            return float(r2_score(predictions, targets))

    def _safe_div(self, arr):
        len_arr = len(arr) - 1
        if len_arr % 10 == 0:
            return np.linspace(0, len_arr, 10).astype(int)
        elif len_arr < 10:
            return np.linspace(0, len_arr, len_arr + 1).astype(int)
        else:
            return np.linspace(0, len_arr, 6).astype(int)

    def _convert_datetime(self, index_data_copy):
        sample_ticks = False
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            if not pd.core.dtypes.common.is_datetime_or_timedelta_dtype(
                index_data_copy
            ):
                try:
                    index_data_copy = pd.to_datetime(index_data_copy)
                except:
                    sample_ticks = True
        return index_data_copy, sample_ticks

    def show_results(self, rows=5):
        """
        Prints the graph with predictions.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer.
                                Number of rows to print.
        =====================   ===========================================
        """
        model = self.learn.model

        model.eval()

        dl = self.learn.data.valid_dl

        targets = []
        predictions = []
        sequence = []
        for i in range(len(dl.x.items)):
            prediction = self._predict(dl.x.items[i])[0]
            target = dl.y.items[i]
            if isinstance(target, (list, np.ndarray)):
                targets.append(target)
            else:
                targets.append([target])
            predictions.append(prediction)
            sequence.append(dl.x.items[i])

        targets = np.array(targets, dtype="float64")
        predictions = np.array(predictions, dtype="float64")

        transformed_results = np.stack(targets, axis=0)
        transformed_results = self._apply_inverse_transform(transformed_results)
        targets_inversed = transformed_results

        transformed_results = np.stack(predictions, axis=0)
        transformed_results = self._apply_inverse_transform(transformed_results)
        predictions_inversed = transformed_results
        column_transforms_mapping = self._data._column_transforms_mapping.copy()
        # del column_transforms_mapping[self._data._dependent_variable]
        sequence_inversed = []
        keys_bk = list(column_transforms_mapping.keys())
        order_columns = (
            self._data._dependent_variable
            + self._data._continuous_variables
            + self._data._categorical_variables
        )
        keys = []
        for key in order_columns:
            if key in keys_bk:
                keys.append(key)

        for seq in sequence:
            seq_inverse = []
            index = 0
            for col_data in seq:
                transformed_data = col_data
                if len(keys) > index:
                    for transform in column_transforms_mapping.get(keys[index]):
                        if isinstance(transform, LabelEncoder):
                            transformed_data = transform.inverse_transform(
                                np.array(transformed_data, dtype=int)
                            )
                        # else: # Commenting it out. Because inverse transform tends to change the scale
                        #     transformed_data = transform.inverse_transform(
                        #         np.array(transformed_data).reshape(-1, 1)
                        #     )
                        #     transformed_data = transformed_data.squeeze(1)

                seq_inverse.append(transformed_data)
                index = index + 1

            sequence_inversed.append(seq_inverse)

        if self._data._index_seq is not None:
            validation_index_seq = self._data._index_seq.take(
                self._data._validation_indexes_ts, axis=0
            )
        else:
            validation_index_seq = None

        import matplotlib.pyplot as plt

        n_items = rows
        if n_items > len(targets_inversed):
            n_items = len(targets_inversed)

        rows = int(n_items)

        fig, axs = plt.subplots(rows, 2, figsize=(10, 10))
        fig.suptitle("Ground truth vs Predictions\n\n", fontsize=16)

        for i in range(rows):
            for idx, seq_plot in enumerate(sequence_inversed[i]):
                sample_ticks = False
                if self._data._index_seq is not None:
                    index_data, sample_ticks = self._convert_datetime(
                        validation_index_seq[i]
                    )
                    axs[i, 0].plot(index_data, seq_plot)
                    axs[i, 1].plot(index_data, seq_plot)
                else:
                    axs[i, 0].plot(seq_plot)
                    axs[i, 1].plot(seq_plot)

                if sample_ticks:
                    axs[i, 0].xaxis.set_major_locator(
                        plt.FixedLocator(
                            self._safe_div(range(len(validation_index_seq[i])))
                        )
                    )
                    axs[i, 1].xaxis.set_major_locator(
                        plt.FixedLocator(
                            self._safe_div(range(len(validation_index_seq[i])))
                        )
                    )
                axs[i, 0].tick_params(axis="x", labelrotation=60)
                axs[i, 1].tick_params(axis="x", labelrotation=60)
                axs[i, 0].set_title(
                    ",".join([f"{val :4f}" for val in targets_inversed[i]])
                )
                axs[i, 1].set_title(
                    ",".join([f"{val :4f}" for val in predictions_inversed[i]])
                )

        plt.tight_layout()
        plt.show()



# model settings
custom_imports = dict(imports=["arcgis.learn.models._prithvi_archs"])

bands = [1, 2, 3, 8, 11, 12]
# required bands Blue, Green, Red, Narrow NIR, SWIR 1, SWIR 2

CLASSES = ("flooded", "non-flooded")

img_norm_flood_model = dict(
    means=[0.14245495, 0.13921481, 0.12434631, 0.31420089, 0.20743526, 0.12046503],
    stds=[0.04036231, 0.04186983, 0.05267646, 0.0822221, 0.06834774, 0.05294205],
)

norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="TemporalEncoderDecoder",
    pretrained=None,
    frozen_backbone=True,
    backbone=dict(
        type="TemporalViTEncoder",
        pretrained=None,
        img_size=224,
        patch_size=16,
        num_frames=1,
        tubelet_size=1,
        in_chans=len(bands),
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4.0,
        norm_pix_loss=False,
    ),
    neck=dict(
        type="ConvTransformerTokensToEmbeddingNeck",
        embed_dim=768,
        output_embed_dim=768,
        drop_cls_token=True,
        Hp=14,
        Wp=14,
    ),
    decode_head=dict(
        num_classes=2,
        in_channels=768,
        ignore_index=2,
        type="FCNHead",
        in_index=-1,
        channels=256,
        num_convs=1,
        concat_input=False,
        dropout_ratio=0.1,
        norm_cfg=dict(type="BN", requires_grad=True),
        align_corners=False,
        loss_decode=dict(
            type="CrossEntropyLoss",
            use_sigmoid=False,
            loss_weight=1,
            class_weight=[0.3, 0.7, 0],
        ),
    ),
    auxiliary_head=dict(
        num_classes=2,
        in_channels=768 * 1,
        ignore_index=2,
        type="FCNHead",
        in_index=-1,
        channels=256,
        num_convs=2,
        concat_input=False,
        dropout_ratio=0.1,
        norm_cfg=dict(type="BN", requires_grad=True),
        align_corners=False,
        loss_decode=dict(
            type="CrossEntropyLoss",
            use_sigmoid=False,
            loss_weight=1,
            class_weight=[0.3, 0.7, 0],
        ),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode="whole"),
)

checkpoint = "https://huggingface.co/ibm-nasa-geospatial/Prithvi-100M-sen1floods11/resolve/main/sen1floods11_Prithvi_100M.pth"



import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

from ._PointRend import (
    generate_regular_grid_point_coords,
    get_uncertain_point_coords_on_grid,
    get_uncertain_point_coords_with_randomness,
    point_sample,
    point_sample_fine_grained_features,
    roi_mask_point_loss,
    c2_msra_fill,
    c2_xavier_fill,
    StandardPointHead,
)

from torchvision.models.detection.roi_heads import (
    RoIHeads,
    fastrcnn_loss,
    maskrcnn_loss,
    maskrcnn_inference,
)

import torch.nn.functional as F
from torch import nn
from torch.jit.annotations import List, Dict


def create_pointrend(model, num_class):
    # get model parameter to create modified RoI head
    y = {}
    y["box_roi_pool"] = model.roi_heads.box_roi_pool
    y["box_head"] = model.roi_heads.box_head
    y["box_predictor"] = model.roi_heads.box_predictor
    y["fg_iou_thresh"] = model.roi_heads.proposal_matcher.high_threshold
    y["bg_iou_thresh"] = model.roi_heads.proposal_matcher.low_threshold
    y["batch_size_per_image"] = model.roi_heads.fg_bg_sampler.batch_size_per_image
    y["positive_fraction"] = model.roi_heads.fg_bg_sampler.positive_fraction
    y["bbox_reg_weights"] = model.roi_heads.box_coder.weights
    y["score_thresh"] = model.roi_heads.score_thresh
    y["nms_thresh"] = model.roi_heads.nms_thresh
    y["detections_per_img"] = model.roi_heads.detections_per_img

    # change mask head to pointrend mask head

    y["mask_roi_pool"] = MaskRoIPoolHead(num_class)
    y["mask_head"] = CoarseMaskHead(num_class, model.backbone.out_channels)
    y["mask_predictor"] = PointRendHeads(
        num_class, in_channel=model.backbone.out_channels
    )

    model.roi_heads = PointRendROIHeads(**y)

    return model


class PointRendROIHeads(RoIHeads):
    def forward(self, features, proposals, image_shapes, targets=None):
        """
        Arguments:
            features (List[Tensor])
            proposals (List[Tensor[N, 4]])
            image_shapes (List[Tuple[H, W]])
            targets (List[Dict])
        """
        train_val = getattr(self, "train_val", False)

        if targets is not None:
            for t in targets:
                floating_point_types = (torch.float, torch.double, torch.half)
                assert (
                    t["boxes"].dtype in floating_point_types
                ), "target boxes must of float type"
                assert (
                    t["labels"].dtype == torch.int64
                ), "target labels must of int64 type"

        if self.training:
            if train_val:
                original_prpsl = [p.clone() for p in proposals]
            (
                proposals,
                matched_idxs,
                labels,
                regression_targets,
            ) = self.select_training_samples(proposals, targets)
        else:
            labels = None
            regression_targets = None
            matched_idxs = None

        box_features = self.box_roi_pool(features, proposals, image_shapes)
        box_features = self.box_head(box_features)
        class_logits, box_regression = self.box_predictor(box_features)

        result = torch.jit.annotate(List[Dict[str, torch.Tensor]], [])
        losses = {}
        if self.training:
            assert labels is not None and regression_targets is not None
            loss_classifier, loss_box_reg = fastrcnn_loss(
                class_logits, box_regression, labels, regression_targets
            )
            losses = {"loss_classifier": loss_classifier, "loss_box_reg": loss_box_reg}
        if not self.training or train_val:
            if train_val:
                box_features = self.box_roi_pool(features, original_prpsl, image_shapes)
                box_features = self.box_head(box_features)
                class_logits, box_regression = self.box_predictor(box_features)
                boxes, scores, labels = self.postprocess_detections(
                    class_logits, box_regression, original_prpsl, image_shapes
                )
            else:
                boxes, scores, labels = self.postprocess_detections(
                    class_logits, box_regression, proposals, image_shapes
                )
            num_images = len(boxes)
            for i in range(num_images):
                result.append(
                    {
                        "boxes": boxes[i],
                        "labels": labels[i],
                        "scores": scores[i],
                    }
                )

        if self.has_mask():
            mask_pred_gt = torch.jit.annotate(List[Dict[str, torch.Tensor]], [])
            if self.training:
                assert matched_idxs is not None
                assert targets is not None
                gt_masks = [t["masks"] for t in targets]
                gt_labels = [t["labels"] for t in targets]

                # during training, only focus on positive boxes
                num_images = len(proposals)
                mask_proposals = []
                pos_matched_idxs = []
                for img_id in range(num_images):
                    pos = torch.nonzero(labels[img_id] > 0).squeeze(1)
                    mask_proposals.append(proposals[img_id][pos])
                    pos_matched_idxs.append(matched_idxs[img_id][pos])
                    mask_pred_gt.append(
                        {
                            "proposal_boxes": proposals[img_id][pos],
                            "gt_classes": gt_labels[img_id][pos_matched_idxs[img_id]],
                            "gt_masks": gt_masks[img_id][pos_matched_idxs[img_id]],
                        }
                    )
            else:
                pos_matched_idxs = None
                for res in result:
                    mask_pred_gt.append(
                        {
                            "pred_boxes": res["boxes"],
                            "pred_classes": res["labels"],
                        }
                    )

            if self.mask_roi_pool is not None:
                mask_features = self.mask_roi_pool(features, mask_pred_gt)
                mask_features = self.mask_head(mask_features)
                mask_logits = self.mask_predictor(features, mask_features, mask_pred_gt)
            else:
                mask_logits = torch.tensor(0)
                raise Exception("Expected mask_roi_pool to be not None")
            loss_mask = {}
            loss_mask_point = {}
            if self.training:
                assert pos_matched_idxs is not None
                assert mask_logits is not None

                rcnn_loss_mask = maskrcnn_loss(
                    mask_features, mask_proposals, gt_masks, gt_labels, pos_matched_idxs
                )
                loss_mask = {"loss_mask": rcnn_loss_mask}
                loss_mask_point = {"loss_mask_point": mask_logits}
            if not self.training or train_val:
                if train_val:
                    mask_pred_gt = []
                    for res in result:
                        mask_pred_gt.append(
                            {
                                "pred_boxes": res["boxes"],
                                "pred_classes": res["labels"],
                            }
                        )
                    mask_features = self.mask_roi_pool.eval()(features, mask_pred_gt)
                    mask_features = self.mask_head.eval()(mask_features)
                    mask_logits = self.mask_predictor.eval()(
                        features, mask_features, mask_pred_gt
                    )
                labels = [r["labels"] for r in result]
                masks_probs = maskrcnn_inference(mask_logits, labels)
                for mask_prob, r in zip(masks_probs, result):
                    r["masks"] = mask_prob

            losses.update(loss_mask)
            losses.update(loss_mask_point)

        return result, losses


def calculate_uncertainty(logits, classes):
    # This code is based on https://github.com/facebookresearch/detectron2/blob/master/projects/PointRend
    """
    We estimate uncerainty as L1 distance between 0.0 and the logit prediction in 'logits' for the
        foreground class in `classes`.

    Args:
        logits (Tensor): A tensor of shape (R, C, ...) or (R, 1, ...) for class-specific or
            class-agnostic, where R is the total number of predicted masks in all images and C is
            the number of foreground classes. The values are logits.
        classes (list): A list of length R that contains either predicted of ground truth class
            for eash predicted mask.

    Returns:
        scores (Tensor): A tensor of shape (R, 1, ...) that contains uncertainty scores with
            the most uncertain locations having the highest uncertainty score.
    """
    if logits.shape[1] == 1:
        gt_class_logits = logits.clone()
    else:
        gt_class_logits = logits[
            torch.arange(logits.shape[0], device=logits.device), classes
        ].unsqueeze(1)
    return -(torch.abs(gt_class_logits))


class MaskRoIPoolHead(nn.Module):
    # This code is based on https://github.com/facebookresearch/detectron2/blob/master/projects/PointRend
    def __init__(self, num_class):
        super().__init__()
        self.mask_coarse_in_features = ["0"]
        self.mask_coarse_side_size = 14
        self._feature_scales = {
            "0": 1.0 / 4,
            "1": 1.0 / 8,
            "2": 1.0 / 16,
            "3": 1.0 / 32,
        }  # FPN block 1/stride

    def forward(self, features, proposals):
        if self.training:
            boxes = [x["proposal_boxes"] for x in proposals]
        else:
            boxes = [x["pred_boxes"] for x in proposals]

        point_coords = generate_regular_grid_point_coords(
            sum(len(x) for x in boxes), self.mask_coarse_side_size, boxes[0].device
        )
        mask_coarse_features_list = [features[k] for k in self.mask_coarse_in_features]
        features_scales = [
            self._feature_scales[k] for k in self.mask_coarse_in_features
        ]
        # For regular grids of points, this function is equivalent to `len(features_list)' calls
        # of `ROIAlign` (with `SAMPLING_RATIO=2`), and concat the results.
        mask_features, _ = point_sample_fine_grained_features(
            mask_coarse_features_list, features_scales, boxes, point_coords
        )

        return mask_features


class CoarseMaskHead(nn.Module):
    # This code is based on https://github.com/facebookresearch/detectron2/blob/master/projects/PointRend
    """
    A mask head with fully connected layers. Given pooled features it first reduces channels and
    spatial dimensions with conv layers and then uses FC layers to predict coarse masks analogously
    to the standard box head.
    """

    def __init__(self, num_class, input_feture_channel):
        """
        The following attributes are parsed from config:
            conv_dim: the output dimension of the conv layers
            fc_dim: the feature dimenstion of the FC layers
            num_fc: the number of FC layers
            output_side_resolution: side resolution of the output square mask prediction
        """
        super(CoarseMaskHead, self).__init__()

        # fmt: off
        self.num_classes            = num_class
        conv_dim                    = 256
        self.fc_dim                 = 1024
        num_fc                      = 2
        self.output_side_resolution = 7
        self.input_channels         = input_feture_channel
        self.input_h                = 14
        self.input_w                = 14
        # fmt: on

        self.conv_layers = []
        if self.input_channels > conv_dim:
            self.reduce_channel_dim_conv = nn.Conv2d(
                self.input_channels,
                conv_dim,
                kernel_size=1,
                stride=1,
                padding=0,
                bias=True,
            )
            self.conv_layers.append(self.reduce_channel_dim_conv)

        self.reduce_spatial_dim_conv = nn.Conv2d(
            conv_dim, conv_dim, kernel_size=2, stride=2, padding=0, bias=True
        )
        self.conv_layers.append(self.reduce_spatial_dim_conv)

        input_dim = conv_dim * self.input_h * self.input_w
        input_dim //= 4

        self.fcs = []
        for k in range(num_fc):
            fc = nn.Linear(input_dim, self.fc_dim)
            self.add_module("coarse_mask_fc{}".format(k + 1), fc)
            self.fcs.append(fc)
            input_dim = self.fc_dim

        output_dim = (
            self.num_classes * self.output_side_resolution * self.output_side_resolution
        )

        self.prediction = nn.Linear(self.fc_dim, output_dim)
        # use normal distribution initialization for mask prediction layer
        nn.init.normal_(self.prediction.weight, std=0.001)
        nn.init.constant_(self.prediction.bias, 0)

        for layer in self.conv_layers:
            c2_msra_fill(layer)
        for layer in self.fcs:
            c2_xavier_fill(layer)

    def forward(self, x):
        # unlike BaseMaskRCNNHead, this head only outputs intermediate
        # features, because the features will be used later by PointHead.
        N = x.shape[0]
        x = x.view(N, self.input_channels, self.input_h, self.input_w)
        for layer in self.conv_layers:
            x = F.relu(layer(x))
        x = torch.flatten(x, start_dim=1)
        for layer in self.fcs:
            x = F.relu(layer(x))
        return self.prediction(x).view(
            N,
            self.num_classes,
            self.output_side_resolution,
            self.output_side_resolution,
        )


class PointRendHeads(torch.nn.Module):
    # This code is based on https://github.com/facebookresearch/detectron2/blob/master/projects/PointRend
    def __init__(self, num_class, **kwargs):
        super().__init__()
        self._feature_scales = {
            "0": 1.0 / 4,
            "1": 1.0 / 8,
            "2": 1.0 / 16,
            "3": 1.0 / 32,
        }  # FPN block 1/stride
        self.mask_point_in_features = ["0"]
        self.mask_point_train_num_points = 14 * 14
        self.mask_point_oversample_ratio = 3
        self.mask_point_importance_sample_ratio = 0.75
        # next two parameters are use in the adaptive subdivions inference procedure
        self.mask_point_subdivision_steps = 5
        self.mask_point_subdivision_num_points = 28 * 28
        in_channel = int(kwargs.get("in_channel", 256))

        in_channels = np.sum([in_channel for f in self.mask_point_in_features])
        self.mask_point_head = StandardPointHead(
            num_class, in_channels, coarse_pred_each_layer=True, fc_dim=in_channel
        )

    def forward(self, features, mask_coarse_logits, instances):
        """
        Forward logic of the mask point head..

        Args:
            features (dict[str, Tensor]): #level input features for mask prediction
            instances (list[Instances]): the per-image instances to train/predict masks.
                In training, they can be the proposals.
                In inference, they can be the predicted boxes.

        Returns:
            In training, a dict of losses.
            In inference, update `instances` with new fields "pred_masks" and return it.
        """

        mask_features_list = [features[k] for k in self.mask_point_in_features]
        features_scales = [self._feature_scales[k] for k in self.mask_point_in_features]

        if self.training:
            proposal_boxes = [x["proposal_boxes"] for x in instances]
            gt_classes = torch.cat([x["gt_classes"] for x in instances])
            with torch.no_grad():
                point_coords = get_uncertain_point_coords_with_randomness(
                    mask_coarse_logits,
                    lambda logits: calculate_uncertainty(logits, gt_classes),
                    self.mask_point_train_num_points,
                    self.mask_point_oversample_ratio,
                    self.mask_point_importance_sample_ratio,
                )

            (
                fine_grained_features,
                point_coords_wrt_image,
            ) = point_sample_fine_grained_features(
                mask_features_list, features_scales, proposal_boxes, point_coords
            )
            coarse_features = point_sample(
                mask_coarse_logits, point_coords, align_corners=False
            )
            point_logits = self.mask_point_head(fine_grained_features, coarse_features)
            loss_mask_point = roi_mask_point_loss(
                point_logits, instances, point_coords_wrt_image
            )
            return loss_mask_point
        else:
            pred_boxes = [x["pred_boxes"] for x in instances]
            pred_classes = torch.cat([x["pred_classes"] for x in instances])
            # The subdivision code will fail with the empty list of boxes
            if len(pred_classes) == 0:
                return mask_coarse_logits

            mask_logits = mask_coarse_logits.clone()
            for subdivions_step in range(self.mask_point_subdivision_steps):
                mask_logits = F.interpolate(
                    mask_logits, scale_factor=2, mode="bilinear", align_corners=False
                )

                # If `mask_point_subdivision_num_points` is larger or equal to the
                # resolution of the next step, then we can skip this step
                H, W = mask_logits.shape[-2:]
                if (
                    self.mask_point_subdivision_num_points >= 4 * H * W
                    and subdivions_step < self.mask_point_subdivision_steps - 1
                ):
                    continue

                uncertainty_map = calculate_uncertainty(mask_logits, pred_classes)
                point_indices, point_coords = get_uncertain_point_coords_on_grid(
                    uncertainty_map, self.mask_point_subdivision_num_points
                )
                fine_grained_features, _ = point_sample_fine_grained_features(
                    mask_features_list, features_scales, pred_boxes, point_coords
                )
                coarse_features = point_sample(
                    mask_coarse_logits, point_coords, align_corners=False
                )
                point_logits = self.mask_point_head(
                    fine_grained_features, coarse_features
                )

                # put mask point predictions to the right places on the upsampled grid.
                R, C, H, W = mask_logits.shape
                point_indices = point_indices.unsqueeze(1).expand(-1, C, -1)
                mask_logits = (
                    mask_logits.reshape(R, C, H * W)
                    .scatter_(2, point_indices, point_logits)
                    .view(R, C, H, W)
                )

            return mask_logits



import types
from torch import nn
from ._arcgis_model import _get_backbone_meta
import logging
import os
import sys
import warnings
import re
import timm

_logger = logging.getLogger(__name__)

try:
    from fastai.callbacks.hooks import hook_outputs, model_sizes
    from fastai.torch_core import one_param
    from fastai.vision import create_body, Image
    from fastai.layers import AdaptiveConcatPool2d
    from fastai.basic_data import DatasetType
    from fastai.callbacks import hook_output
    import torch.nn.functional as F
    from matplotlib import pyplot as plt
    import numpy as np
    import fnmatch
    from timm.models.hub import (
        has_hf_hub,
        load_state_dict_from_hf,
        load_state_dict_from_url,
    )
    from timm.models.helpers import (
        adapt_input_conv,
        build_model_with_cfg,
        overlay_external_default_cfg,
    )
    from torch.hub import get_dir
    import zipfile
    import torch
    from copy import deepcopy

    HAS_FASTAI = True
except Exception as e:
    HAS_FASTAI = False


# same function with modification fastai.vision.learner._test_cnn
def test_cnn_trnsfrmr(m):
    if not isinstance(m, nn.Sequential) or not len(m) == 2:
        return False
    if hasattr(m[1], "_transformer"):
        return True
    return isinstance(m[1][0], (AdaptiveConcatPool2d, nn.AdaptiveAvgPool2d))


def reshape_tensor(x):
    x = x.squeeze()
    if x.ndim > 2:
        return x
    # get feature size(height, width)
    feature_size = int(np.sqrt(x.shape[0]))
    # get number of tokens
    num_tokens = x.shape[0] - feature_size**2
    return x[num_tokens:, :].reshape(feature_size, feature_size, -1).permute(2, 0, 1)


# same function with modification fastai.vision.learner._cl_int_gradcam
def gradcam_trnsfrmr(
    self,
    idx,
    ds_type=None,
    heatmap_thresh=16,
    image=True,
):
    if ds_type == None:
        ds_type = DatasetType.Valid
    m = self.learn.model.eval()
    im, cl = self.learn.data.dl(ds_type).dataset[idx]
    cl = int(cl)
    xb, _ = self.data.one_item(
        im, detach=False, denorm=False
    )  # put into a minibatch of batch size = 1
    with hook_output(m[0]) as hook_a:
        with hook_output(m[0], grad=True) as hook_g:
            preds = m(xb)
            preds[0, int(cl)].backward()
    acts = hook_a.stored[0].cpu()  # activation maps
    grad = hook_g.stored[0][0].cpu()
    if hasattr(m[1], "_transformer"):
        acts = reshape_tensor(acts)
        grad = reshape_tensor(grad)

    if (acts.shape[-1] * acts.shape[-2]) >= heatmap_thresh:
        grad_chan = grad.mean(1).mean(1)
        mult = F.relu(((acts * grad_chan[..., None, None])).sum(0))
        if image:
            xb_im = Image(xb[0])
            _, ax = plt.subplots()
            sz = list(xb_im.shape[-2:])
            xb_im.show(
                ax,
                title=f"pred. class: {self.pred_class[idx]}, actual class: {self.learn.data.classes[cl]}",
            )
            ax.imshow(
                mult,
                alpha=0.4,
                extent=(0, *sz[::-1], 0),
                interpolation="bilinear",
                cmap="magma",
            )
        return mult


hosted_weights = {
    "ecaresnet101d": "682eb6fca513415b9c5a6be61f1bc0f1",
    "ecaresnet101d_pruned": "d676dfa5ffa04732bc8ef58d2c85dd38",
    "ecaresnet269d": "df81aea21521475eba81658496dc6ace",
    "ecaresnet50d": "9f3235316a464885b4d3e5db2d5499b9",
    "ecaresnet50d_pruned": "eca1ccf9ea3145f3ae1cd257f290ef09",
    "ecaresnetlight": "680b9da80da54bbca20a6b7446b5a2df",
    "efficientnet_b1_pruned": "ed49c5e7bb804a05b877702e7ce12f40",
    "efficientnet_b2_pruned": "8eb218d51113449181f1f25efae9ab98",
    "efficientnet_b3_pruned": "b341d13252d244c591370c082bea9696",
    "hardcorenas_a": "af3227e5f48145c2a21c8cf76c3782e4",
    "hardcorenas_b": "9c14f1b482c5465ba5cc75772aeead60",
    "hardcorenas_c": "b758a9d94cb7402c942f2a51fdd99953",
    "hardcorenas_d": "8a927ea1f2c14b2dabaae423c21c193c",
    "hardcorenas_e": "1e7d664452cf4ce596bd5d1bfe776557",
    "hardcorenas_f": "f3a1f74efe244911b166d5456f2bba68",
    "legacy_senet154": "36266e6e22444ce299d76a57a6a817df",
    "legacy_seresnext101_32x4d": "884b2dd7093e49b4884fac2d23bc2386",
    "legacy_seresnext50_32x4d": "e54dc138f33f415984ccfbf6250e1e03",
    "nasnetalarge": "91404c5ecbd842948bb1507f1e313e4b",
    "regnetx_006": "e595c123a67c4a4f87f322b1af4b293c",
    "tf_efficientnet_b6_ns": "20d17115f5db4e11837b8d43e81d59da",
}


# same function with modification timm.models.helpers.load_pretrained
def load_timm_bckbn_pretrained(
    model,
    default_cfg=None,
    num_classes=1000,
    in_chans=3,
    filter_fn=None,
    strict=True,
    progress=True,
):
    default_cfg = default_cfg or getattr(model, "default_cfg", None) or {}
    pretrained_url = default_cfg.get("url", None)
    hf_hub_id = default_cfg.get("hf_hub", None)
    if not pretrained_url and not hf_hub_id:
        _logger.warning(
            "No pretrained weights exist for this model. Using random initialization."
        )
        return

    model_url = hosted_weights.get(default_cfg["architecture"], False)
    if model_url:
        model_dir = os.path.join(get_dir(), "checkpoints")
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        cached_file = os.path.join(model_dir, pretrained_url.split("/")[-1])
        if not os.path.exists(cached_file):
            sys.stderr.write(
                'Downloading: "{}" pretrained weights to {}\n'.format(
                    default_cfg["architecture"], cached_file
                )
            )
            from arcgis.gis import GIS

            gis = GIS(set_active=False)
            item = gis.content.get(model_url)
            item.download(model_dir)
            zipped_file = os.path.join(
                model_dir, pretrained_url.split("/")[-1][:-3] + "zip"
            )
            with zipfile.ZipFile(zipped_file) as f:
                f.extractall(model_dir)
            os.remove(zipped_file)

        state_dict = torch.load(cached_file, map_location="cpu")

    elif hf_hub_id and has_hf_hub(necessary=not pretrained_url):
        _logger.info(f"Loading pretrained weights from Hugging Face hub ({hf_hub_id})")
        state_dict = load_state_dict_from_hf(hf_hub_id)
    else:
        _logger.info(f"Loading pretrained weights from url ({pretrained_url})")
        state_dict = load_state_dict_from_url(
            pretrained_url, progress=progress, map_location="cpu"
        )
    if filter_fn is not None:
        # for backwards compat with filter fn that take one arg, try one first, the two
        try:
            state_dict = filter_fn(state_dict)
        except TypeError:
            state_dict = filter_fn(state_dict, model)

    input_convs = default_cfg.get("first_conv", None)
    if input_convs is not None and in_chans != 3:
        if isinstance(input_convs, str):
            input_convs = (input_convs,)
        for input_conv_name in input_convs:
            weight_name = input_conv_name + ".weight"
            try:
                state_dict[weight_name] = adapt_input_conv(
                    in_chans, state_dict[weight_name]
                )
                _logger.info(
                    f"Converted input conv {input_conv_name} pretrained weights from 3 to {in_chans} channel(s)"
                )
            except NotImplementedError as e:
                del state_dict[weight_name]
                strict = False
                _logger.warning(
                    f"Unable to convert pretrained {input_conv_name} weights, using random init for this layer."
                )

    classifiers = default_cfg.get("classifier", None)
    label_offset = default_cfg.get("label_offset", 0)
    if classifiers is not None:
        if isinstance(classifiers, str):
            classifiers = (classifiers,)
        if num_classes != default_cfg["num_classes"]:
            for classifier_name in classifiers:
                # completely discard fully connected if model num_classes doesn't match pretrained weights
                del state_dict[classifier_name + ".weight"]
                del state_dict[classifier_name + ".bias"]
            strict = False
        elif label_offset > 0:
            for classifier_name in classifiers:
                # special case for pretrained weights with an extra background class in pretrained weights
                classifier_weight = state_dict[classifier_name + ".weight"]
                state_dict[classifier_name + ".weight"] = classifier_weight[
                    label_offset:
                ]
                classifier_bias = state_dict[classifier_name + ".bias"]
                state_dict[classifier_name + ".bias"] = classifier_bias[label_offset:]

    model.load_state_dict(state_dict, strict=strict)


def _default_split(m):
    return (m[1],)


def _tresnet_split(m):
    return (m[0].feature.body.layer4,)


def _squeezenet_split(m):
    return (m[0][0][5], m[0][0][8], m[1])


def _densenet_split(m):
    return (m[0][0][7], m[0][0][9])


def _vgg_split(m):
    return m[0][0][15]


def _rep_vgg(m):
    return m[0][1][2]


def _mobilenetv2_split(m):
    return m[1]


def _darknet_split(m):
    return m[0][1][4]


def _cspres_split(m):
    return m[0][1][3]


def _nfnet_split(m):
    return m[0][1][-1]


def _dpn_split(m):
    return m[0][0][-2]


def _esevovnet_split(m):
    return m[0][1][-1]


def _gernet_split(m):
    return m[0][1][-2]


def _regnet_split(m):
    return m[0][3]


def _resnetv2_split(m):
    return m[0][1][3]


def _modified_cut(m):
    def forward_modified(self, img):
        return self.forward_features(img)

    m.forward = types.MethodType(forward_modified, m)

    class TimmBackbone(nn.Module):
        def __init__(self, m):
            super(TimmBackbone, self).__init__()
            self.feature = m

        def forward(self, x):
            return self.feature.forward_features(x)

    return TimmBackbone(m)


timm_model_meta = {
    "default": {"cut": None, "split": _default_split},
    "squeezenet": {"cut": -1, "split": _squeezenet_split},
    "densenet": {"cut": None, "split": _densenet_split},
    "repvgg": {"cut": -2, "split": _rep_vgg},
    "vgg": {"cut": -2, "split": _vgg_split},
    "mobilenet": {"cut": None, "split": _mobilenetv2_split},
    "darknet": {"cut": None, "split": _darknet_split},
    "hrnet": {"cut": _modified_cut, "split": _default_split},
    "nasnet": {"cut": _modified_cut, "split": _default_split},
    "selecsls": {"cut": _modified_cut, "split": _default_split},
    "tresnet": {"cut": _modified_cut, "split": _tresnet_split},
    "cspres": {"cut": None, "split": _cspres_split},
    "nfnet": {"cut": None, "split": _nfnet_split},
    "dpn": {"cut": None, "split": _dpn_split},
    "ese_vovnet": {"cut": None, "split": _esevovnet_split},
    "gernet": {"cut": None, "split": _gernet_split},
    "nf_regnet": {"cut": None, "split": _nfnet_split},
    "nf_resnet": {"cut": None, "split": _nfnet_split},
    "regnet": {"cut": None, "split": _regnet_split},
    "resnet51q": {"cut": None, "split": _resnetv2_split},
    "resnetv2": {"cut": None, "split": _resnetv2_split},
}


def timm_config(arch):
    model_name = arch if type(arch) is str else arch.__name__
    model_key = [key for key in timm_model_meta if key in model_name] + ["default"]
    return timm_model_meta.get(model_key[0])


def filter_timm_models(flt=[]):
    models = timm.list_models(pretrained=True)
    # remove transformer models
    flt = [
        "*cait*",
        "*coat*",
        "*convit*",
        "*deit*",
        "*gmixer*",
        "*gmlp*",
        "*levit*",
        "*mixer*",
        "*pit*",
        "*resmlp*",
        "*swin*",
        "*tnt*",
        "*twins*",
        "*visformer*",
        "vit_*",
    ] + flt
    flt_models = []
    for f in flt:
        flt_models.extend(fnmatch.filter(models, f))
    return sorted(set(models) - set(flt_models))


def _get_feature_size(arch, cut, chip_size=(64, 64), channel_in=3):
    m = nn.Sequential(*create_body(arch, False, cut).children())
    if "tresnet" in arch.__module__:
        with hook_outputs(m) as hooks:
            dummy_batch = (
                one_param(m)
                .new(1, channel_in, *chip_size)
                .requires_grad_(False)
                .uniform_(-1.0, 1.0)
            )
            x = m.eval()(dummy_batch)
            return [o.stored.shape for o in hooks]
    else:
        return model_sizes(m, chip_size)


def get_backbone(backbone_fn, pretrained):
    if "timm" in backbone_fn.__module__:
        backbone_cut = timm_config(backbone_fn)["cut"]
    elif getattr(backbone_fn, "_is_multispectral", False):
        backbone_cut = _get_backbone_meta(backbone_fn.__name__)["cut"]
    else:
        backbone_cut = None

    return create_body(backbone_fn, pretrained, backbone_cut)


def forward_VisionTransformer(self, x):
    x = self.patch_embed(x)
    cls_token = self.cls_token.expand(x.shape[0], -1, -1)
    if self.dist_token is None:
        x = torch.cat((cls_token, x), dim=1)
    else:
        x = torch.cat((cls_token, self.dist_token.expand(x.shape[0], -1, -1), x), dim=1)
    x = self.pos_drop(x + self.pos_embed)

    x = self.blocks[:-1](x)

    return x


class VisionTransformerHead(nn.Module):
    def __init__(self, block, head, norm, dist_token, pre_logits, head_dist):
        super().__init__()
        self.block = block
        self.head = head
        self.norm = norm
        self.dist_token = dist_token
        self.pre_logits = pre_logits
        self.head_dist = head_dist
        self._transformer = True

    def forward(self, x):
        x = self.block(x)
        x = self.norm(x)
        if self.dist_token is None:
            x = self.pre_logits(x[:, 0])

        if self.head_dist is not None:
            x, x_dist = self.head(x[:, 0]), self.head_dist(x[:, 1])
            return x  # (x + x_dist) / 2
        else:
            x = self.head(x)
        return x


class TransformerClassifierHead(nn.Module):
    def __init__(self, head, head_dist=None):
        super().__init__()
        self.head = head
        self.head_dist = head_dist
        self._transformer = True

    def forward(self, x):
        if self.head_dist is not None:
            x, x_dist = self.head(x[1][:, 0]), self.head_dist(x[1][:, 1])
            return x  # (x + x_dist) / 2
        else:
            x = self.head(x[1])
        return x


def checkpoint_filter_fn_Cait(state_dict, model):
    if "model" in state_dict:
        state_dict = state_dict["model"]
    checkpoint_no_module = {}
    for k, v in state_dict.items():
        new_key = k.replace("module.", "")
        if new_key == "pos_embed" and v.shape != model.pos_embed.shape:
            v = timm.models.vision_transformer.resize_pos_embed(
                v, model.pos_embed, 0, model.patch_embed.grid_size
            )
        checkpoint_no_module[new_key] = v
    return checkpoint_no_module


timm.models.cait.checkpoint_filter_fn = checkpoint_filter_fn_Cait


def forward_Cait(self, x):
    B = x.shape[0]
    x = self.patch_embed(x)
    cls_tokens = self.cls_token.expand(B, -1, -1)
    x = x + self.pos_embed
    x = self.pos_drop(x)

    for i, blk in enumerate(self.blocks):
        x = blk(x)

    for i, blk in enumerate(self.blocks_token_only):
        cls_tokens = blk(x, cls_tokens)

    x_grad_cam = x
    x = torch.cat((cls_tokens, x), dim=1)

    x = self.norm(x)
    return x_grad_cam, x[:, 0]


def forward_CoaT(self, x):
    B = x.shape[0]

    # Serial blocks 1.
    x1 = self.patch_embed1(x)
    H1, W1 = self.patch_embed1.grid_size
    x1 = self.insert_cls(x1, self.cls_token1)
    for blk in self.serial_blocks1:
        x1 = blk(x1, size=(H1, W1))
    x1_nocls = self.remove_cls(x1)
    x1_nocls = x1_nocls.reshape(B, H1, W1, -1).permute(0, 3, 1, 2).contiguous()

    # Serial blocks 2.
    x2 = self.patch_embed2(x1_nocls)
    H2, W2 = self.patch_embed2.grid_size
    x2 = self.insert_cls(x2, self.cls_token2)
    for blk in self.serial_blocks2:
        x2 = blk(x2, size=(H2, W2))
    x2_nocls = self.remove_cls(x2)
    x2_nocls = x2_nocls.reshape(B, H2, W2, -1).permute(0, 3, 1, 2).contiguous()

    # Serial blocks 3.
    x3 = self.patch_embed3(x2_nocls)
    H3, W3 = self.patch_embed3.grid_size
    x3 = self.insert_cls(x3, self.cls_token3)
    for blk in self.serial_blocks3:
        x3 = blk(x3, size=(H3, W3))
    x3_nocls = self.remove_cls(x3)
    x3_nocls = x3_nocls.reshape(B, H3, W3, -1).permute(0, 3, 1, 2).contiguous()

    # Serial blocks 4.
    x4 = self.patch_embed4(x3_nocls)
    H4, W4 = self.patch_embed4.grid_size
    x4 = self.insert_cls(x4, self.cls_token4)
    for blk in self.serial_blocks4[:-1]:
        x4 = blk(x4, size=(H4, W4))
    x_grad_cam = x4
    x4 = self.serial_blocks4[-1](x4, size=(H4, W4))
    x4_nocls = self.remove_cls(x4)
    x4_nocls = x4_nocls.reshape(B, H4, W4, -1).permute(0, 3, 1, 2).contiguous()

    if self.parallel_blocks is None:
        x4 = self.norm4(x4)
        x4_cls = x4[:, 0]
        return x_grad_cam, x4_cls

    # Parallel blocks.
    for blk in self.parallel_blocks:
        x2, x3, x4 = (
            self.cpe2(x2, (H2, W2)),
            self.cpe3(x3, (H3, W3)),
            self.cpe4(x4, (H4, W4)),
        )
        x1, x2, x3, x4 = blk(
            x1, x2, x3, x4, sizes=[(H1, W1), (H2, W2), (H3, W3), (H4, W4)]
        )

    x2 = self.norm2(x2)
    x3 = self.norm3(x3)
    x4 = self.norm4(x4)
    x2_cls = x2[:, :1]  # [B, 1, C]
    x3_cls = x3[:, :1]
    x4_cls = x4[:, :1]
    merged_cls = torch.cat((x2_cls, x3_cls, x4_cls), dim=1)  # [B, 3, C]
    merged_cls = self.aggregate(merged_cls).squeeze(dim=1)  # Shape: [B, C]

    return x_grad_cam, merged_cls


def checkpoint_filter_fn_Convit(state_dict, model):
    if "model" in state_dict:
        state_dict = state_dict["model"]
    checkpoint_no_module = {}
    for k, v in state_dict.items():
        if k == "pos_embed" and v.shape != model.pos_embed.shape:
            v = timm.models.vision_transformer.resize_pos_embed(
                v, model.pos_embed, 0, model.patch_embed.grid_size
            )
        checkpoint_no_module[k] = v
    return checkpoint_no_module


def create_convit(variant, pretrained=False, **kwargs):
    if kwargs.get("features_only", None):
        raise RuntimeError(
            "features_only not implemented for Vision Transformer models."
        )

    return build_model_with_cfg(
        timm.models.convit.ConViT,
        variant,
        pretrained,
        default_cfg=timm.models.convit.default_cfgs[variant],
        pretrained_filter_fn=checkpoint_filter_fn_Convit,
        **kwargs,
    )


timm.models.convit._create_convit = create_convit


def forward_ConViT(self, x):
    B = x.shape[0]
    x = self.patch_embed(x)

    cls_tokens = self.cls_token.expand(B, -1, -1)

    if self.use_pos_embed:
        x = x + self.pos_embed
    x = self.pos_drop(x)

    no_of_block = len(self.blocks)
    for u, blk in enumerate(self.blocks):
        if u == self.local_up_to_layer:
            x = torch.cat((cls_tokens, x), dim=1)
        if u + 1 == no_of_block - 1:
            x_grad_cam = x
        x = blk(x)

    x = self.norm(x)
    return x_grad_cam, x[:, 0]


def checkpoint_filter_fn_levit(state_dict, model):
    if "model" in state_dict:
        state_dict = state_dict["model"]
    D = model.state_dict()
    for k in state_dict.keys():
        if "attention_bias" in k:
            if "attention_biases" in k and D[k].shape != state_dict[k].shape:
                state_dict[k] = F.interpolate(
                    state_dict[k][None][None], size=D[k].shape, mode="bilinear"
                ).squeeze()
            else:
                state_dict[k] = D[k]
        if k in D and D[k].ndim == 4 and state_dict[k].ndim == 2:
            state_dict[k] = state_dict[k][:, :, None, None]
    return state_dict


timm.models.levit.checkpoint_filter_fn = checkpoint_filter_fn_levit


def forward_Levit(self, x):
    x = self.patch_embed(x)
    if not self.use_conv:
        x = x.flatten(2).transpose(1, 2)
    x = self.blocks(x)
    x_grad_cam = x
    x = x.mean((-2, -1)) if self.use_conv else x.mean(1)
    # concat same cls_token in dim 1 for distillation head
    x = torch.cat((x.unsqueeze(1), x.unsqueeze(1)), dim=1)
    return x_grad_cam, x


def checkpoint_filter_fn_pit(state_dict, model):
    out_dict = {}
    p_blocks = re.compile(r"pools\.(\d)\.")
    for k, v in state_dict.items():
        if k == "pos_embed" and v.shape != model.pos_embed.shape:
            v = F.interpolate(v, size=model.pos_embed.shape[-2:], mode="bilinear")
        k = p_blocks.sub(lambda exp: f"transformers.{int(exp.group(1))}.pool.", k)
        out_dict[k] = v
    return out_dict


timm.models.pit.checkpoint_filter_fn = checkpoint_filter_fn_pit


def forward_pit_transformer(self, x):
    x, cls_tokens = x
    B, C, H, W = x.shape
    token_length = cls_tokens.shape[1]

    x = x.flatten(2).transpose(1, 2)
    x = torch.cat((cls_tokens, x), dim=1)

    x = self.blocks[:-1](x)
    x_grad_cam = x
    x = self.blocks[-1](x)

    cls_tokens = x[:, :token_length]
    x = x[:, token_length:]
    x = x.transpose(1, 2).reshape(B, C, H, W)

    if self.pool is not None:
        x, cls_tokens = self.pool(x, cls_tokens)
    return x_grad_cam, cls_tokens


def forward_PoolingVisionTransformer(self, x):
    x = self.patch_embed(x)
    x = self.pos_drop(x + self.pos_embed)
    cls_tokens = self.cls_token.expand(x.shape[0], -1, -1)
    x, cls_tokens = self.transformers((x, cls_tokens))
    cls_tokens = self.norm(cls_tokens)
    if self.head_dist is not None:
        return x, cls_tokens
    return x, cls_tokens[:, 0]


def checkpoint_filter_fn_swin(state_dict, model):
    """convert patch embedding weight from manual patchify + linear proj to conv"""
    out_dict = {}
    if "model" in state_dict:
        state_dict = state_dict["model"]
    for k, v in state_dict.items():
        if "patch_embed.proj.weight" in k and len(v.shape) < 4:
            O, I, H, W = model.patch_embed.proj.weight.shape
            v = v.reshape(O, -1, H, W)
        elif "relative_position" in k or "attn_mask" in k:
            # To resize embedding when using model at different size from pretrained weights
            model_attr = k.split(".")
            temp_module = model
            for attr in model_attr:
                # last temp_module will be tensor
                temp_module = temp_module.__getattr__(attr)
            if v.shape == temp_module.shape:
                continue
            if "index" in model_attr[-1]:
                v = temp_module
            elif "attn_mask" in model_attr[-1]:
                v = temp_module
            elif "bias_table" in model_attr[-1]:
                v = F.interpolate(
                    v[None][None], size=list(temp_module.shape), mode="bilinear"
                ).squeeze()

        out_dict[k] = v
    return out_dict


def create_swin_transformer(variant, pretrained=False, default_cfg=None, **kwargs):
    if default_cfg is None:
        default_cfg = deepcopy(timm.models.swin_transformer.default_cfgs[variant])
    overlay_external_default_cfg(default_cfg, kwargs)
    default_num_classes = default_cfg["num_classes"]
    default_img_size = default_cfg["input_size"][-2:]

    num_classes = kwargs.pop("num_classes", default_num_classes)
    img_size = kwargs.pop("img_size", default_img_size)
    if kwargs.get("features_only", None):
        raise RuntimeError(
            "features_only not implemented for Vision Transformer models."
        )
    kwargs["window_size"] = img_size // 32

    model = build_model_with_cfg(
        timm.models.swin_transformer.SwinTransformer,
        variant,
        pretrained,
        default_cfg=default_cfg,
        img_size=img_size,
        num_classes=num_classes,
        pretrained_filter_fn=checkpoint_filter_fn_swin,
        **kwargs,
    )

    return model


timm.models.swin_transformer._create_swin_transformer = create_swin_transformer


def forward_SwinTransformer(self, x):
    x = self.patch_embed(x)
    if self.absolute_pos_embed is not None:
        x = x + self.absolute_pos_embed
    x = self.pos_drop(x)
    x = self.layers(x)
    x_grad_cam = x
    x = self.norm(x)  # B L C
    x = self.avgpool(x.transpose(1, 2))  # B C 1
    x = torch.flatten(x, 1)
    return x_grad_cam, x


def forward_TNT(self, x):
    B = x.shape[0]
    pixel_embed = self.pixel_embed(x, self.pixel_pos)

    patch_embed = self.norm2_proj(
        self.proj(self.norm1_proj(pixel_embed.reshape(B, self.num_patches, -1)))
    )
    patch_embed = torch.cat((self.cls_token.expand(B, -1, -1), patch_embed), dim=1)
    patch_embed = patch_embed + self.patch_pos
    patch_embed = self.pos_drop(patch_embed)

    no_of_block = len(self.blocks)
    for idx, blk in enumerate(self.blocks):
        pixel_embed, patch_embed = blk(pixel_embed, patch_embed)
        if idx + 1 == no_of_block - 1:
            x_grad_cam = patch_embed

    patch_embed = self.norm(patch_embed)
    return x_grad_cam, patch_embed[:, 0]


def forward_Twins(self, x):
    B = x.shape[0]
    for i, (embed, drop, blocks, pos_blk) in enumerate(
        zip(self.patch_embeds, self.pos_drops, self.blocks, self.pos_block)
    ):
        x, size = embed(x)
        x = drop(x)
        for j, blk in enumerate(blocks):
            x = blk(x, size)
            if j == 0:
                x = pos_blk(x, size)  # PEG here
        if i < len(self.depths) - 1:
            x = x.reshape(B, *size, -1).permute(0, 3, 1, 2).contiguous()
    x_grad_cam = x
    x = self.norm(x)
    return x_grad_cam, x.mean(dim=1)  # GAP here


def checkpoint_filter_fn_visformer(state_dict, model):
    out_dict = {}
    D = model.state_dict()
    for k, v in state_dict.items():
        if "pos_embed" in k and v.shape != D[k].shape:
            v = F.interpolate(v, size=D[k].shape[-2:], mode="bilinear")
        out_dict[k] = v
    return out_dict


def create_visformer(variant, pretrained=False, default_cfg=None, **kwargs):
    if kwargs.get("features_only", None):
        raise RuntimeError(
            "features_only not implemented for Vision Transformer models."
        )
    model = build_model_with_cfg(
        timm.models.visformer.Visformer,
        variant,
        pretrained,
        default_cfg=timm.models.visformer.default_cfgs[variant],
        pretrained_filter_fn=checkpoint_filter_fn_visformer,
        **kwargs,
    )
    return model


timm.models.visformer._create_visformer = create_visformer


def forward_Visformer(self, x):
    if self.stem is not None:
        x = self.stem(x)

    # stage 1
    x = self.patch_embed1(x)
    if self.pos_embed:
        x = x + self.pos_embed1
        x = self.pos_drop(x)
    for b in self.stage1:
        x = b(x)

    # stage 2
    if not self.vit_stem:
        x = self.patch_embed2(x)
        if self.pos_embed:
            x = x + self.pos_embed2
            x = self.pos_drop(x)
    for b in self.stage2:
        x = b(x)

    # stage3
    if not self.vit_stem:
        x = self.patch_embed3(x)
        if self.pos_embed:
            x = x + self.pos_embed3
            x = self.pos_drop(x)
    for b in self.stage3:
        x = b(x)

    x_grad_cam = x
    x = self.norm(x)
    x = self.global_pool(x)

    return x_grad_cam, x


def get_transformer_backbone(backbone_name, num_classes, img_size, pretrained):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        logging.disable(logging.WARNING)
        transformer_backbone = timm.create_model(
            backbone_name,
            num_classes=num_classes,
            img_size=img_size,
            pretrained=pretrained,
        )
        logging.disable(logging.NOTSET)

    model_name = transformer_backbone.__class__.__name__
    forward_function = globals()["forward_" + model_name]
    transformer_backbone.forward = types.MethodType(
        forward_function, transformer_backbone
    )
    return transformer_backbone


def create_transformer_FeatureClassifier(
    backbone_name, num_classes, img_size, pretrained
):
    transformer_backbone = get_transformer_backbone(
        backbone_name, num_classes, img_size, pretrained
    )

    model_name = transformer_backbone.__class__.__name__

    if model_name == "VisionTransformer":
        trnsfrmr_head = VisionTransformerHead(
            transformer_backbone.blocks[-1],
            transformer_backbone.head,
            transformer_backbone.norm,
            transformer_backbone.dist_token,
            transformer_backbone.pre_logits,
            transformer_backbone.head_dist,
        )
    else:
        head_dist = getattr(transformer_backbone, "head_dist", None)
        trnsfrmr_head = TransformerClassifierHead(transformer_backbone.head, head_dist)

    if model_name == "PoolingVisionTransformer":
        transformer_backbone.transformers[-1].forward = types.MethodType(
            forward_pit_transformer, transformer_backbone.transformers[-1]
        )

    transformer_model = nn.Sequential(transformer_backbone, trnsfrmr_head)
    return transformer_model


def shortened_transformer_backbone():
    shortened_name = [
        "cait_tiny_24",
        "cait_tiny_36",
        "cait_small_24",
        "cait_small_36",
        "cait_base_36",
        "cait_base_48",
        "deit_tiny",
        "deit_small",
        "deit_base",
        "pit_tiny",
        "pit_mini",
        "pit_small",
        "pit_base",
        "swin_tiny_window7",
        "swin_small_window7",
        "swin_base_window7",
        "swin_base_window12",
        "swin_large_window7",
        "swin_large_window12",
        "tnt_small",
        "vit_tiny",
        "vit_small",
        "vit_base",
        "vit_large",
        "vit_huge",
    ]
    return shortened_name


# backbone with unique image sizes
def map_transformer_backbone_unique(backbone):
    transformer_backbone_unique = {
        "cait_small_36": "cait_s36_384",
        "cait_base_36": "cait_m36_384",
        "cait_base_48": "cait_m48_448",
        "deit_tiny": "deit_tiny_patch16_224",
        "deit_small": "deit_small_patch16_224",
        "pit_tiny": "pit_ti_224",
        "pit_mini": "pit_xs_224",
        "pit_small": "pit_s_224",
        "pit_base": "pit_b_224",
        "swin_tiny_window7": "swin_tiny_patch4_window7_224",
        "swin_small_window7": "swin_small_patch4_window7_224",
        "swin_base_window7": "swin_base_patch4_window7_224",
        "swin_base_window12": "swin_base_patch4_window12_384",
        "swin_large_window7": "swin_large_patch4_window7_224",
        "swin_large_window12": "swin_large_patch4_window12_384",
        "tnt_small": "tnt_s_patch16_224",
        "vit_huge": "vit_huge_patch14_224_in21k",
    }
    return transformer_backbone_unique.get(backbone, None)


# backbone with different image sizes
def map_transformer_backbone(backbone):
    transformer_backbone = {
        "cait_tiny_24": "cait_xxs24",
        "cait_tiny_36": "cait_xxs36",
        "cait_small_24": "cait_s24",
        "deit_base": "deit_base_patch16",
        "vit_tiny": "vit_tiny_patch16",
        "vit_small": "vit_small_patch16",
        "vit_base": "vit_base_patch16",
        "vit_large": "vit_large_patch16",
    }

    return transformer_backbone.get(backbone, None)


def complete_transformer_backbone_name(backbone, img_size):
    if backbone is not None and "timm:" in backbone:
        bckbn = backbone.split(":")[1]
        img_size = 224 if abs(img_size - 224) < abs(img_size - 384) else 384
        shortened_backbone = shortened_transformer_backbone()
        if bckbn in shortened_backbone:
            if map_transformer_backbone_unique(bckbn) is not None:
                return f"timm:{map_transformer_backbone_unique(bckbn)}"
            bckbn = map_transformer_backbone(bckbn)
            return f"timm:{bckbn}_{img_size}"
    return backbone



"""
arcgis.realtime.Velocity provides API functions to automate the ArcGIS Velocity REST API.
ArcGIS Velocity is a real-time and big data processing and analysis capability of ArcGIS Online.
It enables you to ingest, visualize, analyze, store, and act upon data from Internet of Things (IoT) sensors.

"""
from .velocity import Velocity

from .velocity import FeedsManager
from .velocity import RealTimeAnalyticsManager
from .velocity import BigDataAnalyticsManager

from .feeds_manager import Feed
from .realtime_analytics_manager import RealTimeAnalytics
from .bigdata_analytics_manager import BigDataAnalytics
from .http_authentication_type import NoAuth, CertificateAuth, BasicAuth
from ._reserved_fields import _ReservedFields
